import os
import re
import math
import sys
from argparse import Namespace
from .writers import write_gff3
import PhiSpyRepeatFinder

def find_repeat(fn, st, ppno, extraDNA, output_dir):
    if len(fn) == 0:
        print("Len sequence is 0 so ignoring\n")
        return {}

    rep = {}
    index = 0

    try:
        repeats = PhiSpyRepeatFinder.repeatFinder(fn, 3)
    except Exception as e:
        sys.stderr.write("There was an error running repeatfinder for {}:{}\n".format(fn, e))
        return {}

    for r in repeats:
        if (r['first_start'] < (3 * extraDNA)) and (r['second_start'] > (len(fn) - (3 * extraDNA))):
            # check that start is always less than end
            # This always causes an off by one error, so we have to increment our ends
            if r['first_end'] < r['first_start']:
                [r['first_start'], r['first_end']] = [r['first_end'] + 1, r['first_start'] + 1]
            if r['second_end'] < r['second_start']:
                [r['second_start'], r['second_end']] = [r['second_end'] + 1, r['second_start'] + 1]

            rep[index] = {}
            rep[index]['s1'] = r['first_start'] + st
            rep[index]['e1'] = r['first_end'] + st
            rep[index]['s2'] = r['second_start'] + st
            rep[index]['e2'] = r['second_end'] + st
            index += 1

    return rep

def check_intg(prophage_sta,prophage_sto,rep,integ,con):
    for m in integ:
        if integ[m]['contig'] != con:
            continue
        if integ[m]['start'] < (prophage_sta + (prophage_sto - prophage_sta) / 2):
            l = integ[m]['start']
        else:
            l = integ[m]['stop']
        if (l - rep['s1'] <= 500 and l - rep['s1'] > 0) or (l - rep['e1'] <= 500 and l - rep['e1'] > 0):
            return 1
        if (l - rep['s2'] <= 500 and l - rep['s2'] > 0) or (l - rep['e2'] <= 500 and l - rep['e2'] > 0):
            return 1
    return 0

def find_smallest(a,b):
    mm = 1000000
    for i in a:
        for j in b:
            if math.fabs(i - j) < mm:
                mm = math.fabs(i - j)
    return mm

def find_rna(prophage_start, prophage_stop, repeat_list, record, cont, integrs):
    #try:
    #    infile = open(organism_path + '/Features/rna/tbl', 'r')
    #except:
    #    sys.exit('Cannot open ' + organism_path + "/Features/rna/tbl")
    my_start = 1000000
    start_end = 0
    end_start = 0
    my_end = 1000000
    mydiff = 1000000
    #for line in infile:
    for feature in record.get_entry(cont).get_features('tRNA'):
        #temp = re.split('\t', line.strip())
        #if len(temp) < 3:
        #    continue
        #if 'trna' in temp[2].lower() or 'tmrna' in temp[2].lower():
        #    if ',' in temp[1]:
        #        ttemp = re.split(',', temp[1])
        #        temp[1] = ttemp[len(ttemp) - 1]
        #    temp1 = re.split('_', temp[1])
        #    #contig = temp1[len(temp1)-3]
        #    contig = temp[1][:temp[1][:temp[1].rfind('_')].rfind('_')]
        #    start = int(temp1[len(temp1) - 2])
        #    stop = int(temp1[len(temp1) - 1])
        #    if cont == contig:
        i = 0
        while i < len(repeat_list):
            a = find_smallest([feature.start, feature.stop], 
                              [repeat_list[i]['s1'], repeat_list[i]['s2'],
                               repeat_list[i]['e1'], repeat_list[i]['e2']])
            if check_intg(prophage_start, prophage_stop, repeat_list[i], integrs, cont) == 1:
                if (math.fabs(repeat_list[i]['s1'] - repeat_list[i]['e2']) > math.fabs(
                        my_start - my_end)) or mydiff == 1000000:
                    my_start = repeat_list[i]['s1']
                    my_end = repeat_list[i]['e2']
                    start_end = repeat_list[i]['e1']
                    end_start = repeat_list[i]['s2']
                    mydiff = a
            if a <= 500 and a < mydiff:
                my_start = repeat_list[i]['s1']
                my_end = repeat_list[i]['e2']
                start_end = repeat_list[i]['e1']
                end_start = repeat_list[i]['s2']
                mydiff = a
            i += 1
    #infile.close()
    if mydiff == 1000000:
        return '0_0'  # 'null'
    return str(my_start) + '_' + str(my_end)+'_'+ str(start_end)+'_'+ str(end_start)

def check_pp(contig,start,stop,pp):
    if start > stop:
        (start, stop) = (stop, start)

    for j in pp:
        if contig == pp[j]['contig']:
            if pp[j]['start'] <= start and pp[j]['stop'] >= stop:
                return j
    return 0

def check_phage_word_start(sjcontig, a, b, c):
    j = 0
    tot = 0
    for i in c:
        start = c[i]['start']
        stop = c[i]['stop']
        if start > stop:
            t = start
            start = stop
            stop = t
        if a <= start and stop <= b and c[i]['contig'] == sjcontig:
            if c[i]['pp'] > 0.5:
                j += 1
            tot += 1
    if tot < 4 * j:
        return a
    else:
        return b

def check_phage_word_end(sjcontig, a, b, c):
    j = 0
    tot = 0
    for i in c:
        start = c[i]['start']
        stop = c[i]['stop']
        if start > stop:
            t = start
            start = stop
            stop = t
        if a <= start and stop <= b and c[i]['contig'] == sjcontig:
            if c[i]['pp'] > 0.5:
                j = j + 1
            tot = tot + 1
    if tot < 4 * j:
        return b
    else:
        return a
               
def final_check_phage_word(sjcontig,a,b,c):
    j = 0
    tot = 0
    for i in c:
        start = c[i]['start']
        stop = c[i]['stop']
        if start > stop:
            t = start
            start = stop
            stop = t
        if a <= start and stop <= b and c[i]['contig'] == sjcontig:
            if c[i]['pp'] > 0:
                j = j + 1
            tot = tot + 1
    if j > 5 and tot < 2 * j:
        return str(a) + '_' + str(b)
    else:
        return '0_0'

def clarification_by_phage_word(sjcontig, bef_start, bef_stop, aft_start, aft_stop, genome):
    if aft_start == 0 and aft_stop == 0:
        return '0_0'
    if bef_start <= aft_start:
        s = check_phage_word_start(sjcontig, bef_start, aft_start, genome)
    else:
        s = check_phage_word_start(sjcontig, aft_start, bef_start, genome)

    if bef_stop <= aft_stop:
        e = check_phage_word_end(sjcontig, bef_stop, aft_stop, genome)
    else:
        e = check_phage_word_end(sjcontig, aft_stop, bef_stop, genome)
    se = final_check_phage_word(sjcontig, s, e, genome)
    return se

def fixing_start_end(**kwargs): #output_dir, organism_path, INSTALLATION_DIR, phageWindowSize, non_prophage_gene_gaps=10):
    self = Namespace(**kwargs)
    try:
        infile = open(os.path.join(self.output_dir, 'initial_tbl.tsv'), 'r')
    except:
        sys.exit('ERROR: Cannot open initial_tbl.txt in fixing_start_end')

    #make all predicted pp list
    print("Checking prophages in initial_tbl.tsv\n")
    pp = {}
    i = 0
    flag = 0
    intg = {}
    intg_index = 1
    genome = {}
    index = 1
    temp = {}
    distance_from_last_prophage = 1000
    for line in infile:
        oldtemp = temp
        temp = line.strip().split("\t")
        distance_from_last_prophage += 1
        if temp[1] == 'function':
            continue
        # Find location of all prophage regions
        if float(temp[7]) >= 1 or float(temp[8]) >= 1:
            # This is a prophage region. Is it a new prophage
            new_prophage = False
            # check the sequences are on the same contig. If not, definitely a new prophage
            if temp[2] != oldtemp[2]:
                new_prophage = True
            if flag == 0 and distance_from_last_prophage > self.nonprophage_genegaps:
                # we need at least 10 non phage genes betweeen prophages. This should be a variable.
                new_prophage = True
            if new_prophage:
                i += 1
                pp[i] = {}
                pp[i]['contig'] = temp[2]
                pp[i]['start'] = min(int(temp[3]), int(temp[4]))
                pp[i]['stop'] = max(int(temp[3]), int(temp[4]))
                pp[i]['num genes'] = 1
                flag = 1
            else:
                pp[i]['stop'] = max(pp[i]['stop'], int(temp[3]), int(temp[4]))
                pp[i]['num genes'] += 1
            distance_from_last_prophage = 0
        else:
            if temp[0] == 'fig|160490.1.peg.707':
                sys.stderr.write("BUGGGER: Got to here but shouldn't\n")
            flag = 0
            # Find location of integrases
        if float(temp[8]) == 1.5:
            intg[intg_index] = {}
            intg[intg_index]['start'] = min(int(temp[3]), int(temp[4]))
            intg[intg_index]['stop'] = max(int(temp[3]), int(temp[4]))
            intg[intg_index]['contig'] = str(temp[2])
            intg_index += 1
        genome[index] = {}
        genome[index]['start'] = int(temp[3])
        genome[index]['stop'] = int(temp[4])
        genome[index]['pp'] = float(temp[8])
        genome[index]['contig'] = temp[2]
        genome[index]['function'] = temp[1]
        genome[index]['rank'] = float(temp[6])
        index += 1
    infile.close()
    #######################################################################################
    #                                                                                     #
    # Filter the potential prophages based on how many potential prophage genes are       #
    # in the window, and tell us which ones were dropped.                                 #
    #                                                                                     #
    #######################################################################################
    temppp = {}
    j = 1
    prophagesummary = []
    for i in pp:
        if pp[i]['num genes'] >= self.number:
            temppp[j] = pp[i]
            j += 1
            prophagesummary.append([pp[i]['contig'], pp[i]['start'], pp[i]['stop'], pp[i]['num genes'], "Kept"])
        else:
            prophagesummary.append([pp[i]['contig'], pp[i]['start'], pp[i]['stop'], pp[i]['num genes'], "Dropped. Not enough genes"])
    # print a list of all prophages and the number of genes
    if prophagesummary:
        sys.stderr.write('Potential prophages (sorted highest to lowest)\n')
        sys.stderr.write('Contig\tStart\tStop\tNumber of potential genes\tStatus\n')
        for p in sorted(prophagesummary, key=lambda x: x[3], reverse=True):
            sys.stderr.write("\t".join(map(str, p)) + "\n")
    pp = temppp
    sys.stderr.write("\n")
    # End filtering
    # find start end for all pp using repeat finder
    #dna = read_contig(organism_path)
    dna = {entry.id : str(entry.seq) for entry in self.record}
    extraDNA = 2000
    for i in pp:
        print("PROPHAGE: " + str(i) + " Contig: " + str(pp[i]['contig']) + " Start: " + str(
            pp[i]['start']) + " Stop: " + str(pp[i]['stop']))
        start = pp[i]['start'] - extraDNA
        if start < 1:
            start = 1
        if 'stop' in pp[i]:
            stop = pp[i]['stop'] + extraDNA
        else:
            stop = genome[len(genome) - 1]['stop']
        if (stop > len(dna[pp[i]['contig']])):
            stop = len(dna[pp[i]['contig']])
        if stop - start > 200000:
            print("Not checking repeats for pp " + str(i) + " because it is too big: " + str(stop - start) + "\n")
            continue
        sys.stderr.write("PP: " + str(i) + " start: " + str(pp[i]['start']) + " stop: " + str(pp[i]['stop']) + "\n")
        print("Finding repeats in pp " + str(i) + " from " + str(start) + " to " + str(stop))
        repeat_list = find_repeat(dna[pp[i]['contig']][start:stop], start, i, extraDNA, self.output_dir)
        s_e = find_rna(start, stop, repeat_list, self.record, pp[i]['contig'], intg)
        if s_e != 'null':
            t = re.split('_', s_e)
            if s_e == '0_0':
                s_e1 = clarification_by_phage_word(pp[i]['contig'], pp[i]['start'], pp[i]['stop'], pp[i]['start'],
                                                   pp[i]['stop'], genome)
            else:
                s_e1 = clarification_by_phage_word(pp[i]['contig'], pp[i]['start'], pp[i]['stop'], float(t[0]),
                                                   float(t[1]), genome)
            t1 = re.split('_', s_e1)
            # we only accept this if it increases the phage length
            if float(t1[0]) > 0 and float(t1[0]) < pp[i]['start'] and float(t1[1]) > 0 and float(t1[1]) > pp[i]['stop']:
                pp[i]['start'] = float(t1[0])
                pp[i]['stop'] = float(t1[1])
                sys.stderr.write("\tReset start of prophage " + str(i) + " to " + str(pp[i]['start']) + " and stop to ")
                sys.stderr.write(str(pp[i]['stop']) + " after checking phage words\n")

            if (float(t[0]) != 0) and (pp[i]['start'] == float(t[0])) and (pp[i]['stop'] == float(t[1])):
                #1 if (float(t[0])!= 0) and (float(t1[0]) == float(t[0])) and (float(t1[1]) == float(t[1])):
                temps1 = min(t[0], t[2])
                tempe1 = max(t[0], t[2])
                temps2 = min(t[1], t[3])
                tempe2 = max(t[1], t[3])
                pp[i]['att'] = [
                    t[0],
                    t[2], 
                    t[3],
                    t[1], 
                    dna[pp[i]['contig']][int(temps1) - 1:int(tempe1)],
                    dna[pp[i]['contig']][int(temps2) - 1:int(tempe2)],
                    'Repeat exactly at the end'
                ]
                # string representation of the above
                pp[i]['atts'] = "\t".join(map(str, pp[i]['att']))
            else:
                # this approach will just append the longest repeat
                longestrep = 0
                bestrep = None
                samelenrep = 0
                for idx in repeat_list:
                    lengthrep = math.fabs(repeat_list[idx]['e1'] - repeat_list[idx]['s1'])
                    if (lengthrep > longestrep) and (lengthrep < 150):
                        longestrep = lengthrep
                        bestrep = repeat_list[idx]
                        samelenrep = 1
                    elif (lengthrep == longestrep):
                        samelenrep += 1
                if bestrep:
                    attLseq = dna[pp[i]['contig']][int(bestrep['s1']) - 1:int(bestrep['e1']) - 1]
                    attRseq = dna[pp[i]['contig']][int(bestrep['s2']) - 1:int(bestrep['e2']) - 1]
                    if len(attLseq) == 0:
                        print("Could not get a sequence from " + str(int(bestrep['s1']) - 1) + " to " + str(
                            int(bestrep['e1']) - 1) + " from " + str(pp[i]['contig']) + " (length: " + str(
                            len(dna[pp[i]['contig']])) + ")\n")
                    if len(attRseq) == 0:
                        print("Could not get a sequence from " + str(int(bestrep['s2']) - 1) + " to " + str(
                            int(bestrep['e2']) - 1) + " from " + str(pp[i]['contig']) + " (length: " + str(
                            len(dna[pp[i]['contig']])) + ")\n")
                    pp[i]['att'] = [
                        bestrep['s1'],
                        bestrep['e1'],
                        bestrep['s2'],
                        bestrep['e2'],
                        attLseq,
                        attRseq,
                        "Longest Repeat flanking phage and within " + str(extraDNA) + " bp"
                    ]
                    pp[i]['atts'] = "\t".join(map(str, pp[i]['att']))
                    if samelenrep > 1:
                        sys.stderr.write("There were {} repeats with the same length as the best. One chosen somewhat randomly!\n".format(samelenrep))
    # fix start end for all pp
    try:
        infile = open(os.path.join(self.output_dir, 'initial_tbl.tsv'), 'r')
        outfile = open(os.path.join(self.output_dir, 'prophage_tbl.tsv'), 'w')
    except:
        sys.exit('ERROR: Cannot open initial_tbl.tsv')

    for line in infile:
        temp = re.split('\t', line.strip())
        if temp[1] == 'function':
            outfile.write(line)
            continue
        me = check_pp(temp[2], int(temp[3]), int(temp[4]), pp)
        if me == 0:
            outfile.write(line.strip() + '\t0' + '\n')
        else:
            outfile.write(line.strip() + '\t' + str(me) + '\n')
    infile.close()
    outfile.close()
    os.remove(os.path.join(self.output_dir, 'initial_tbl.tsv'))
    # print the prophage coordinates:
    out = open(os.path.join(self.output_dir, 'prophage_coordinates.tsv'), 'w')
    for i in pp:
        if 'atts' not in pp[i]:
            pp[i]['atts']=""
        out.write("\t".join(map(str, ["pp" + str(i), "", pp[i]['contig'], pp[i]['start'], pp[i]['stop'], pp[i]['atts']])) + "\n")
    out.close()
    # write the prophage location table
    out = open(os.path.join(self.output_dir, "prophage.tbl"), "w")
    for i in pp:
        out.write("pp_" + str(i) + "\t" + str(pp[i]['contig']) + "_" + str(pp[i]['start']) + "_" + str(pp[i]['stop']) + "\n")
    out.close()

    # write the prophage in GFF3 format
    write_gff3(self.output_dir, pp)

def make_prophage_tbl(inputf, outputf):
    try:
        f = open(inputf, 'r')
        fw = open(outputf, 'w')
    except:
        print('Cant open', inputf, ' or ', outputf)
        return
    pp = {}
    ppindx = 0
    prev_contig = None
    inphage = False
    header = f.readline()
    for line in f:
        temp = line.strip().split("\t")
        if int(temp[9]) > 0:
            newphage = False
            if temp[2] != prev_contig:
                newphage = True
            if not inphage:
                newphage = True
            if newphage:
                ppindx += 1
                pp[ppindx] = {}
                pp[ppindx]['contig'] = temp[2]
                pp[ppindx]['start'] = min(int(temp[3]), int(temp[4]))
                pp[ppindx]['stop'] = max(int(temp[3]), int(temp[4]))
            else:
                pp[ppindx]['stop'] = max(int(temp[3]), int(temp[4]))
            inphage = True
            prev_contig = temp[2]
        else:
            inphage = False
    for i in pp:
        fw.write("pp_" + str(i) + "\t" + pp[i]['contig'] + "_" + str(pp[i]['start']) + "_" + str(pp[i]['stop']) + "\n")
    f.close()
    fw.close()

################################################################################
#def call_start_end_fix(output_dir, organismPath, INSTALLATION_DIR, threshold_for_FN, phageWindowSize):
    # Make the prophage_tbl_temp.txt file.
    #fixing_start_end(output_dir,organismPath,INSTALLATION_DIR)
    #fixing_start_end(output_dir, organismPath, INSTALLATION_DIR, phageWindowSize)
    #make_prophage_tbl(output_dir + 'prophage_tbl.tsv', output_dir + 'prophage.tbl')
    #fixing_false_negative(output_dir, threshold_for_FN, phageWindowSize)
    # Make the prophage_tbl_temp.txt file.
    #make_prophage_tbl(output_dir+'prophage_tbl.txt',output_dir+'prophage.tbl')



# -*- coding: utf-8 -*-
import os
import sys
import argparse
import re
import pkg_resources
from argparse import RawTextHelpFormatter
from argparse import ArgumentTypeError as err
from .pathtype import PathType


def print_list():
    f = None
    try:
        # with pip we use resource streams that may be files or from archives
        f = pkg_resources.resource_stream('PhiSpyModules', 'data/trainingGenome_list.txt')
    except:
        sys.stderr.write('cannot find list')
        sys.exit(-1)
    for line in f:
        line = line.decode().strip()
        temp = re.split('\t', line)
        if int(temp[3]) == 1:
            print("{}\t{}".format(temp[2], 'data/' + temp[1]))
    f.close()

def is_valid_file(x):
    if not x or not os.path.exists(x):
        raise argparse.ArgumentTypeError("Checking for validity: {0} does not exist".format(x))
    return x


def get_args():
    usage = 'python3 PhiSpy.py [-opt1, [-opt2, ...]] infile'
    parser = argparse.ArgumentParser(
        description="phiSpy is a program for identifying prophages from among microbial genome sequences",
        epilog="(c) 2008-2018 Sajia Akhter, Katelyn McNair, Przemysław Decewicz, Rob Edwards, San Diego State University, San Diego, CA")
    parser.add_argument('infile', type=is_valid_file, help='Input file in genbank format', nargs='?')
    parser.add_argument('-m', '--make_training_data', type=str,
                             help='Create training data from a set of annotated genome files. Requires is_phage=1 qualifier in prophage\'s CDSs')
    parser.add_argument('-t', '--training_set', action='store', default='data/trainSet_genericAll.txt',
                             help='Choose the most closely related set to your genome.')
    parser.add_argument('-l', '--list', action='store_true', default=False,
                             help='List the available training sets and exit')
    #parser.add_argument('-c', '--choose', type=bool, default=False, const=True, nargs='?',
    #                         help='Choose a training set from a list (overrides -t)')
    parser.add_argument('-e', '--evaluate', type=bool, default=False, const=True, nargs='?',
                             help='Run in evaluation mode -- does not generate new data, but reruns the evaluation')
    parser.add_argument('-n', '--number', default=5, type=int,
                             help='Number of consecutive genes in a region of window size that must be prophage genes to be called. [Default: 5]')
    parser.add_argument('-w', '--window_size', default=30, type=int,
                             help='Window size of consecutive genes to look through to find phages. [Default: 30]')
    parser.add_argument('-g', '--nonprophage_genegaps', default=10, type=int,
                             help='The number of non phage genes betweeen prophages. [Default: 10]')
    parser.add_argument('-r', '--randomforest_trees', default=500, type=int,
                             help='Number of trees generated by Random Forest classifier. [Default: 500]')
    parser.add_argument('--expand_slope', action='store_true', default=False, 
                             help='Use the product of the slope of the Shannon scores in making test sets')
    parser.add_argument('--kmers_type', default='all', choices=['all', 'codon', 'simple'], type=str,
                             help='Type of kmers used for calculating Shannon scores. [Default: all]')
    #parser.add_argument('-i', '--input_dir', help='The input directory that holds the genome')
    parser.add_argument('-o', '--output_dir', help='The output directory to write the results')
    parser.add_argument('-qt', '--quiet', type=bool, default=False, const=True, nargs='?',
                             help='Run in quiet mode')
    parser.add_argument('-k', '--keep', type=bool, default=False, const=True, nargs='?',
                             help='Do not delete temp files')
    parser.add_argument('-v', '--version', type=bool, default=False, const=True, nargs='?', help="Print the version and exit")
    return parser.parse_args()



import types


class SeqioFilter( list ):
    """This class is to allow filtering of the Biopython SeqIO record

    SeqIO.parse returns a generator object so anytime you want to perform
    an action on it, you must iterate through the entire list. This
    class add the ability to filter and return only a subset of the
    features.

    Note:
        To use simply pass a SeqIO.parse object to it and then when
        the object is called a keyword is passed to it and only those
        features matching the keyword are returned.
    Example:
        record = SeqioFilter(SeqIO.parse(infile)):
        #no change to standard SeqIO calls
        for entry in record:
            print(entry.id, entry.seq)
        #now we can get only certain features
        for cds in record.get_feature('CDS'):
            print(cds)

    """
    def __init__( self, content ):
        self.n = 0
        self.get_n = dict()
        for n, item in enumerate(content):
            self.attach_methods(item)
            self.get_n[item.id] = n
            self.append(item)

    def __iter__(self):
        return self

    def __next__(self):
        try:
            item = self[self.n]
        except IndexError:
            self.n = 0
            raise StopIteration()
        self.n += 1
        return item

    def __call__( self, keyword='' ):
        pass

    def get_entry(self, id):
        return self[self.get_n[id]]
 
    def attach_methods(self, target):
        """This method allows attaching new methods to the SeqIO entry object

           Args:
               target: is the SeqIO object that will be attaching a method to
        """
        def get_features(target, feature_type):
            for feature in target.features:
                feature.id       = " ".join(feature.qualifiers.get('locus_tag', [str(feature.location)]))
                feature.function = " ".join(feature.qualifiers.get('product',['unknown']))
                feature.start    = int(feature.location.start) + 1
                feature.stop     = int(feature.location.end)
                if feature.strand < 0:
                    feature.start, feature.stop = feature.stop, feature.start
                if not feature_type or feature.type == feature_type:
                    yield feature
        target.get_features = types.MethodType(get_features,target)




#!/usr/bin/env python
import os
import sys
import subprocess
import types

from Bio import SeqIO

INSTALLATION_DIR = os.path.dirname(os.path.realpath(__file__)) + '/'
sys.path.append(INSTALLATION_DIR)

import PhiSpyModules


def main(argv):  #organismPath, output_dir, trainingFlag, INSTALLATION_DIR, evaluateOnly, threshold_for_FN, phageWindowSize, quietMode, keep):

    ######################################
    #         parse the options          #
    ######################################
    args_parser = PhiSpyModules.get_args()

    ######################################
    #   list the training sets and exit  #
    ######################################
    if args_parser.list:
        PhiSpyModules.print_list()
        exit(0)

    if args_parser.version:
        print("PhiSpy version: {}".format(PhiSpyModules.version.__version__))
        sys.exit(0)

    # if we get here we need an input file
    if not args_parser.infile:
        sys.stderr.write("ERROR: Please provide an input file. Use -h for more options\n")
        sys.exit(-1)

    # in future support other types
    input_file = PhiSpyModules.SeqioFilter(SeqIO.parse(args_parser.infile, "genbank"))
    args_parser.record = input_file
    if not args_parser.output_dir:
        sys.stderr.write("ERROR: Output directory (-o) is required\n")
        sys.exit(-1)
    os.makedirs(args_parser.output_dir, exist_ok=True)

    ######################################
    #         make training set          #
    ######################################
    if args_parser.make_training_data:
        print('Making Train Set...')
        my_make_train_flag = PhiSpyModules.make_set_train(**vars(args_parser))
        exit()

    ######################################
    #         make testing set           #
    ######################################
    print('Making Test Set...')
    my_make_test_flag = PhiSpyModules.make_test_set(**vars(args_parser))
    # check file im,plement later
    #if (my_make_test_flag == 0):
    #    print('The input organism is too small to predict prophages. Please consider large contig (having at least 40 genes) to use PhiSpy.')
    #    return

    ######################################
    #         do classification          #
    ######################################
    print('Start Classification Algorithm...')
    PhiSpyModules.call_randomforest(**vars(args_parser))
    PhiSpyModules.make_initial_tbl(**vars(args_parser))

    ######################################
    #         i dont know what           #
    ######################################
    ###### added in this version 2.2 #####
    if (args_parser.training_set == 'data/trainSet_genericAll.txt'):
        print('As training flag is zero, considering unknown functions')
        PhiSpyModules.consider_unknown(args_parser.output_dir)

    ######################################
    #         do evaluation              #
    ######################################
    print('Evaluating...')
    PhiSpyModules.fixing_start_end(**vars(args_parser))
    print('Done!!!')



if __name__== "__main__":
    main(sys.argv)



"""
A module to write the output in different formats
"""

import os
import sys
__author__ = 'Rob Edwards'


def write_gff3(output_dir, pp, verbose=False):
    """
    Write GFF3 code. This was adapted from code contribued by [Jose Francisco Sanchez-Herrero]
    (https://github.com/JFsanchezherrero/)

    :param output_dir: the location to write to
    :param pp: the list of prophage objects
    :param verbose: more output
    """

    ## GFF output
    out_GFF = open(os.path.join(output_dir, "prophage.gff3"), "w")
    out_GFF.write("##gff-version 3")
    out_GFF.write("\n")
    
    ## loop through pp results
    for i in pp:
        ## list
        list_write = ["pp" + str(i), str(pp[i]['contig']), str(pp[i]['start']), str(pp[i]['stop'])]

        ## GFF  
        ## strand is not known...
        out_GFF.write(str(pp[i]['contig']) + 
                      '\tPhiSpy\tprophage_region\t' +
                      str(pp[i]['start']) + '\t' + 
                      str(pp[i]['stop']) + '\t.\t.' + 
                      ' \t.\tID=pp' + str(i))
        out_GFF.write('\n')

        ## is there att sequence?
        if 'att' not in pp[i]:
            pp[i]['att']=""
        else:
            ## attL 
            out_GFF.write(str(pp[i]['contig']) +
                          '\tPhiSpy\tattL\t' +
                          str(pp[i]['att'][0]) + '\t' + 
                          str(pp[i]['att'][1]) + '\t.\t.' + 
                          ' \t.\tID=pp' + str(i))
            out_GFF.write('\n')

            ## attR 
            out_GFF.write(str(pp[i]['contig']) +
                          '\tPhiSpy\tattR\t' +
                          str(pp[i]['att'][2]) + '\t' +
                          str(pp[i]['att'][3]) + '\t.\t.' +
                          ' \t.\tID=pp' + str(i))
            out_GFF.write('\n')
        
    out_GFF.close()




import re
import string
import os

def unknown_func(x):
    x_lower = x.lower()
    if (
       (len(x) == 0) or
       #('hypoth' in x_lower) or
       ('conserved protein' in x_lower) or
       ('gene product' in x_lower) or
       ('interpro' in x_lower) or
       ('uncharacterized' in x_lower) or
       ('pseudogene' in x_lower) or
       ('similar to' in x_lower) or
       ('similarity' in x_lower) or
       ('glimmer' in x_lower) or
       ('unknown' in x_lower) or
       ('complete' in x_lower) or
       ('ensang' in x_lower) or
       ('unnamed' in x_lower) or
       ('Expressed' in x_lower) or
       ('similar to' in x_lower) or
       (' identi' in x_lower) or
       ('ortholog of' in x_lower) or
       ('structural feature' in x_lower) or
       ('cds_' in x_lower) or
       ('predicted by Psort' in x) or
       ('AGR_' in x) or
       ('EG:' in x) or
       ('RIKEN' in x) or
       re.search('lmo\d+ protein', x_lower) or
       re.search('lmo\d+protein', x_lower) or
       re.search('B[sl][lr]\d', x_lower) or
       re.search('^U\d', x) or
       re.search('[a-zA-Z]{2,3}\|', x) or
       re.search('orf\d+', x_lower) or
       re.match('orf[^_]', x_lower) or
       re.match('predicted', x_lower) or
       re.match('bh\d+', x_lower) or
       re.match('y[a-z]{2,4}\\b', x) or
       re.match('[a-z]{2,3}\d+[^:\+\-0-9]', x_lower) ):
        return 1
    else:
        return 0

def add_unknown_function_initial_tbl(infile,outfile):
    try:
        f = open(infile,'r')
        fw = open(outfile,'w')
    except:
        print('ERROR: Cannot open initial_tbl.tsv add_unknown_function_initial_tbl.')
        return 0
    flag = 0
    for line in f:
        if flag == 0:
            fw.write(line)
            flag = 1
            continue
        line = line.strip()
        temp = re.split('\t',line)
        i = 0
        while i<8:
            fw.write(temp[i]+'\t')
            i = i + 1
        x = unknown_func(temp[1])
        if x == 0:
            fw.write(temp[8]+'\n')
        else:
            fw.write('0.5\n')
    f.close()
    fw.close()
    return 1

def consider_unknown(output_dir):
    it = os.path.join(output_dir, 'initial_tbl.tsv')
    it2 = os.path.join(output_dir, 'initial_tbl_2.tsv')
    x = add_unknown_function_initial_tbl(it, it2)
    if (x == 1):
        cmd2 = "rm " + it
        os.system(cmd2)
        cmd2 = "mv " + it2 + ' ' + it
        os.system(cmd2)



# read and set the version

from .seqio_filter import SeqioFilter
from .makeTrain import make_set_train
from .makeTest import make_test_set
from .classification import call_randomforest, make_initial_tbl
from .unknownFunction import consider_unknown
from .evaluation import fixing_start_end
from .helper_functions import get_args, print_list
from .pathtype import PathType
from .writers import write_gff3

from .version import __version__

__all__ = ['SeqioFilter', 'make_set_train', 'make_test_set',
           'call_randomforest', 'make_initial_tbl', 'consider_unknown',
           'fixing_start_end', 'get_args', 'print_list', 'PathType',
           'write_gff3'
           ]



import re
import sys
import string
import os
import pkg_resources
from io import TextIOWrapper
from sklearn.cluster import KMeans
from sklearn.ensemble import RandomForestClassifier
import numpy as np
from argparse import Namespace

def find_training_genome(trainingFlag, INSTALLATION_DIR):
    try:
        # f = open(os.path.join(INSTALLATION_DIR, 'data/trainingGenome_list.txt'), 'r')
        f = pkg_resources.resource_stream('PhiSpyModules', 'data/trainingGenome_list.txt')
    except:
        print('cannot open resource stream training genomes at data/trainingGenome_list.txt')
        return ''

    for line in f:
        temp = re.split('\t',line.decode().strip())
        if int(temp[0]) == trainingFlag:
            f.close()
            return temp[1].strip()
    return ''

def call_randomforest(**kwargs):
    output_dir = kwargs['output_dir']
    trainingFile = kwargs['training_set']
    infile = os.path.join(output_dir, "testSet.txt")
    outfile = os.path.join(output_dir, "classify.tsv")
    #train_data = np.genfromtxt(fname=trainingFile, delimiter="\t", skip_header=1, filling_values=1) # why not fill missing values with 0?
    # convert this to run from pip
    if not pkg_resources.resource_exists('PhiSpyModules', trainingFile):
        sys.stderr.write("FATAL: Can not find data file {}\n".format(trainingFile))
        sys.exit(-1)
    strm = pkg_resources.resource_stream('PhiSpyModules', trainingFile)
    train_data = np.genfromtxt(TextIOWrapper(strm), delimiter="\t", skip_header=1, filling_values=1)
    test_data = np.genfromtxt(fname=infile, delimiter="\t", skip_header=1, filling_values=1)
    # Przemek's comment
    # by default 10 until version 0.22 where default is 100
    # number of estimators also implies the precision of probabilities, generally 1/n_estimators
    # in R's randomForest it's 500 and the usage note regarding number of trees to grow says:
    # "This should not be set to too small a number, to ensure that every input row gets predicted at least a few times."
    clf = RandomForestClassifier(n_estimators = kwargs['randomforest_trees'])
    clf.fit(train_data[:, :-1], train_data[:, -1].astype('int'))
    np.savetxt(outfile, clf.predict_proba(test_data)[:,1])

def my_sort(orf_list):
     n = len(orf_list)
     i = 1
     while( i <= n ):
          j = i + 1
          while( j < n ):
               flag = 0
               #direction for both
               if( orf_list[i]['start'] < orf_list[i]['stop'] ):
                    dir_i = 1
               else:
                    dir_i = -1
               if( orf_list[j]['start'] < orf_list[j]['stop'] ):
                    dir_j = 1
               else:
                    dir_j = -1

               #check whether swap need or not
               if dir_i == dir_j:
                    if orf_list[i]['start']>orf_list[j]['start']:
                         flag = 1
               else:
                    if dir_i == 1:
                         if orf_list[i]['start']>orf_list[j]['stop']:
                              flag = 1
                    else:
                         if orf_list[i]['stop']>orf_list[j]['start']:
                              flag = 1
               #swap
               if flag == 1:
                    temp = orf_list[i]
                    orf_list[i] = orf_list[j]
                    orf_list[j] = temp
               j = j+1
          i = i+1
     return orf_list

def find_mean(all_len):
     sum = 0.0
     for i in all_len:
          sum = sum + i
     return float(sum)/len(all_len)

def calc_pp(func):
    x = 0
    func = func.replace('-',' ')
    func = func.replace(',',' ')
    a = re.split(' ',func)
    if (
       'phage'         in a or
       'lysin'         in a or
       'endolysin'     in a or
       'holin'         in a or
       'capsid'        in a or
       'tail'          in a or
       'bacteriophage' in a or
       'prophage'      in a or
       'portal'        in a or
       'terminase'     in a or
       'tapemeasure'   in a or
       'baseplate'     in a or 
       'virion'        in a or
       'antirepressor' in a or
       'excisionase'   in a or
       'mobile element protein' == func or
       re.search(r"\b%s\b" % "tape measure", func) or
       re.search(r"\b%s\b" % "Cro-like repressor", func) or
       re.search(r"\b%s\b" % "CI-like repressor", func) or
       re.search(r"\b%s\b" % "rIIA lysis", func) or
       re.search(r"\b%s\b" % "rI lysis", func) or
       re.search(r"\b%s\b" % "rIIB lysis", func) or
       re.search(r"\b%s\b" % "base plate", func)
       ):
           x = 1
    elif ('unknown' in func) or ('hypothetical' in func):
           x = 0.5
    else:
           x = 0
    if 'recombinase' in func or 'integrase' in func:
           x = 1.5
    if ('phage' in func) and ('shock' in func):
           x = 0
    if "dna binding domain" in func:
           x = 0
    return x

def calc_function_3files(organism):
    my_func = {}
    x = 0 #no need it for computation.. just a flag in "except"
    try:
        f_fun = open(organism+'/proposed_non_ff_functions','r')
        for line in f_fun:
            temp = re.split('\t',line.strip())
            if len(temp)>=2:
                my_func[temp[0]] = temp[1]
        f_fun.close()
    except:
        x = x + 1
    try:
        f_fun = open(organism+'/proposed_functions','r')
        for line in f_fun:
            temp = re.split('\t',line.strip())
            if len(temp)>=2:
                my_func[temp[0]] = temp[1]
        f_fun.close()
    except:
        x = x + 1
    try:
        f_fun = open(organism+'/assigned_functions','r')
        for line in f_fun:
            temp = re.split('\t',line.strip())
            if len(temp)>=2:
                my_func[temp[0]] = temp[1]
        f_fun.close()
    except:
        x = x + 1

    return my_func

def input_bactpp(**kwargs):
    print(kwargs)
    exit()
   
    #bact_file = organism+'/Features/peg/tbl'
    #try:
    #    fh = open(bact_file,'r')
    #except:
    #    print('cant open file- assigned functions/tbl file:',organism)
    #    return {}
    #my_func = calc_function_3files(organism)
    all_orf_list = {}
    for i in fh:
        temp = re.split('\t',i.strip())
        temp1 = re.split('_',temp[1])
        if ',' in temp[1]:
            ttemp = re.split(',',temp[1])
            temp[1] = ttemp[len(ttemp)-1]
        temp1 = re.split('_',temp[1])
        contig = temp[1][:temp[1][:temp[1].rfind('_')].rfind('_')]
        start = int(temp1[len(temp1)-2])
        stop = int(temp1[len(temp1)-1])
        #save info for sorting orf
        if contig in all_orf_list:
            x = len(all_orf_list[contig]) + 1
        else:
            x = 1
            all_orf_list[contig]={}
        all_orf_list[contig][x]={}
        all_orf_list[contig][x]['fig'] = temp[0]
        all_orf_list[contig][x]['contig'] = str(contig)
        all_orf_list[contig][x]['start'] = start
        all_orf_list[contig][x]['stop'] = stop
        if temp[0] in my_func:
            all_orf_list[contig][x]['function'] = my_func[temp[0]]
            all_orf_list[contig][x]['pp'] = calc_pp(my_func[temp[0]].lower(),INSTALLATION_DIR)
        else:
            all_orf_list[contig][x]['function'] = "-"
            all_orf_list[contig][x]['pp'] = 0.5
    fh.close()
    all = {}
    index = 1
    for mycontig in all_orf_list:
        orf_list = my_sort(all_orf_list[mycontig])
        i = 1
        while i <= len(orf_list):
            all[index] = {}
            all[index]['fig'] = orf_list[i]['fig']
            all[index]['function'] = orf_list[i]['function']
            all[index]['contig'] = orf_list[i]['contig']
            all[index]['start'] = orf_list[i]['start']
            all[index]['stop'] = orf_list[i]['stop']
            all[index]['rank'] = 0.0
            all[index]['status'] = 0
            all[index]['pp'] = orf_list[i]['pp']
            i = i+1
            index = index+1
    return all

def make_initial_tbl(**kwargs): #organismPath, output_dir, window, INSTALLATION_DIR):
    self = Namespace(**kwargs)
    x = []
    for entry in self.record:
        for feature in entry.get_features('CDS'):
            all = {}
            all['fig'] = feature.id
            all['function'] = feature.function
            all['contig'] = entry.id
            all['start'] = feature.start
            all['stop'] = feature.stop
            all['rank'] = 0.0
            all['status'] = 0
            all['pp'] = calc_pp(feature.function)
            x.append(all)
    try:
        infile = open(os.path.join(self.output_dir, 'classify.tsv'), 'r')
        outfile = open(os.path.join(self.output_dir, 'initial_tbl.tsv'), 'w')
    except:
        sys.exit('ERROR: Cannot open classify.tsv in make_initial_tbl')
    #x = input_bactpp(**kwargs)
    j = 1
    ranks = [[] for n in range(len(x))]
    for line in infile:
        val = float(line.strip())
        for k in range(j-int(self.window_size/2), j+int(self.window_size/2)):
            if k <= 0 or k >= len(x) or j >= len(x) or x[k]['contig'] != x[j]['contig']:
                continue
            ranks[k].append(val)
        j += 1
    infile.close()
    #calculate threshold
    y = []
    j = 1
    while j < len(x):
        x[j]['rank'] = sum(ranks[j]) / len(ranks[j]) 
        x[j]['extra'] =  ranks[j] 
        y.append(x[j]['rank'])
        #y.append([x[j]['rank']])
        j = j+1
    #threshold = max(y)/2
    y2 = np.array(y).reshape(-1, 1)
    km = KMeans(n_clusters = 2)
    km.fit(y2)
    centers = km.cluster_centers_
    threshold = max(centers[0][0], centers[1][0])
    """
    Note added by Rob:
    At this point we have the classifications for each ORF and we want to take a sliding window and decide where the phage should
    start. We have two calculations for a threshold for the rank: either the kmeans centers and finding things above the larger center
    or just a plain threshold.
    
    """
    j = 1
    outfile.write('fig_no\tfunction\tcontig\tstart\tstop\tposition\trank\tmy_status\tpp\tFinal_status\tstart of attL\tend of attL\tstart of attR\tend of attR\tsequence of attL\tsequence of attR\tReason for att site\n')
    while j < len(x):
        if x[j]['rank'] > threshold:
            x[j]['status'] = 1
        outfile.write(str(x[j]['fig']))
        outfile.write('\t')
        outfile.write(str(x[j]['function']))
        outfile.write('\t')
        outfile.write(str(x[j]['contig']))
        outfile.write('\t')
        outfile.write(str(x[j]['start']))
        outfile.write('\t')
        outfile.write(str(x[j]['stop']))
        outfile.write('\t')
        outfile.write(str(j))
        outfile.write('\t')
        outfile.write(str(x[j]['rank']))
        outfile.write('\t')
        outfile.write(str(x[j]['status']))
        outfile.write('\t')
        outfile.write(str(x[j]['pp']))
        outfile.write('\n')
        j = j+1
    outfile.close()
